@@ -288,3 +288,179 @@ macro_rules! opaque_identifier {
         partial_eq_string!(Box<$id>);
     };
 }
+
+macro_rules! opaque_identifier_validated {
+    (
+        $( #[doc = $docs:literal] )*
+        $vis:vis type $id:ident [ $validate_id:ident ];
+    ) => {
+        $( #[doc = $docs] )*
+        #[repr(transparent)]
+        #[derive(PartialEq, Eq, PartialOrd, Ord, Hash)]
+        #[cfg_attr(feature = "serde", derive(serde::Serialize), serde(transparent, crate = "serde"))]
+        pub struct $id(str);
+
+        paste::paste! {
+            doc_concat! {
+                #[doc = concat!("An owned [", stringify!($id), "].")]
+                pub type [<$id Box>] = Box<$id>;
+            }
+        }
+
+        impl $id {
+            #[allow(clippy::transmute_ptr_to_ptr)]
+            fn from_borrowed(s: &str) -> &Self {
+                unsafe { std::mem::transmute(s) }
+            }
+
+            pub(super) fn from_owned(s: Box<str>) -> Box<Self> {
+                unsafe { Box::from_raw(Box::into_raw(s) as _) }
+            }
+
+            fn into_owned(self: Box<Self>) -> Box<str> {
+                unsafe { Box::from_raw(Box::into_raw(self) as _) }
+            }
+
+            doc_concat! {
+                #[doc = concat!("Creates a string slice from this `", stringify!($id), "`.")]
+                pub fn as_str(&self) -> &str {
+                    &self.0
+                }
+            }
+
+            doc_concat! {
+                #[doc = concat!("Creates a byte slice from this `", stringify!($id), "`.")]
+                pub fn as_bytes(&self) -> &[u8] {
+                    self.0.as_bytes()
+                }
+            }
+        }
+
+        impl std::fmt::Debug for $id {
+            fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
+                f.write_str(self.as_ref())
+            }
+        }
+
+        impl Clone for Box<$id> {
+            fn clone(&self) -> Self {
+                (**self).to_owned()
+            }
+        }
+
+        impl ToOwned for $id {
+            type Owned = Box<$id>;
+
+            fn to_owned(&self) -> Self::Owned {
+                Self::from_owned(self.0.into())
+            }
+        }
+
+        impl From<&$id> for Box<$id> {
+            fn from(id: &$id) -> Self {
+                id.to_owned()
+            }
+        }
+
+        impl AsRef<str> for $id {
+            fn as_ref(&self) -> &str {
+                self.as_str()
+            }
+        }
+
+        impl AsRef<str> for Box<$id> {
+            fn as_ref(&self) -> &str {
+                self.as_str()
+            }
+        }
+
+        impl From<&$id> for std::rc::Rc<$id> {
+            fn from(s: &$id) -> std::rc::Rc<$id> {
+                let rc = std::rc::Rc::<str>::from(s.as_str());
+                unsafe { std::rc::Rc::from_raw(std::rc::Rc::into_raw(rc) as *const $id) }
+            }
+        }
+
+        impl From<&$id> for std::sync::Arc<$id> {
+            fn from(s: &$id) -> std::sync::Arc<$id> {
+                let arc = std::sync::Arc::<str>::from(s.as_str());
+                unsafe { std::sync::Arc::from_raw(std::sync::Arc::into_raw(arc) as *const $id) }
+            }
+        }
+
+
+
+        impl From<Box<$id>> for String {
+            fn from(id: Box<$id>) -> Self {
+                id.into_owned().into()
+            }
+        }
+
+        #[cfg(feature = "serde")]
+        impl<'de> serde::Deserialize<'de> for Box<$id> {
+            fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
+            where
+                D: serde::Deserializer<'de>,
+            {
+                use serde::de::Error;
+
+                let s = String::deserialize(deserializer)?;
+
+                match try_from(s) {
+                    Ok(o) => Ok(o),
+                    Err(e) => Err(D::Error::custom(e)),
+                }
+            }
+        }
+
+        fn try_from<S>(s: S) -> Result<Box<$id>, crate::Error>
+        where
+            S: AsRef<str> + Into<Box<str>>,
+        {
+            $validate_id(s.as_ref())?;
+            Ok($id::from_owned(s.into()))
+        }
+
+        impl<'a> std::convert::TryFrom<&'a str> for &'a $id {
+            type Error = crate::Error;
+
+            fn try_from(s: &'a str) -> Result<Self, Self::Error> {
+                $validate_id(s)?;
+                Ok($id::from_borrowed(s))
+            }
+        }
+
+        impl std::str::FromStr for Box<$id> {
+            type Err = crate::Error;
+
+            fn from_str(s: &str) -> Result<Self, Self::Err> {
+                try_from(s)
+            }
+        }
+
+        impl std::convert::TryFrom<&str> for Box<$id> {
+            type Error = crate::Error;
+
+            fn try_from(s: &str) -> Result<Self, Self::Error> {
+                try_from(s)
+            }
+        }
+
+        impl std::convert::TryFrom<String> for Box<$id> {
+            type Error = crate::Error;
+
+            fn try_from(s: String) -> Result<Self, Self::Error> {
+                try_from(s)
+            }
+        }
+
+        impl std::fmt::Display for $id {
+            fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
+                write!(f, "{}", self.as_str())
+            }
+        }
+
+        partial_eq_string!($id);
+        partial_eq_string!(Box<$id>);
+    }
+}
